/*
 * TOYO CC/PCI PCI Interface to CC/7700 Controller
 * 
 * Hidetada Baba
 * baba@ribf.riken.jp
 * last modified : 07/03/01 15:27:24 
 */

#include "cc7700.h"

void cc77_init_register(int irq,int ior,int pbor);
int cc77_check_done(int try);
int cc77_check_emp_fifo(void);
int cc77_check_full_fifo(void);
int cc77_check_lam(void);
void cc77_read_lam(int *data);
int cc77_control(int cnaf);
int cc77_read16(int cnaf,short *data);
int cc77_read24(int cnaf,short *data);
int cc77_write16(int cnaf,short *data);
int cc77_write24(int cnaf,int *data);
int cc77_block_read16(int mode,int cnaf,short *data,int count);
int cc77_block_read24(int mode,int cnaf,unsigned int *data,int count);
int cc77_dma_block_read16(int mode,int cnaf,int count);
int cc77_dma_block_read24(int mode,int cnaf,int count);
void cc77_crate_reset(void);
void cc77_rfs_enable_interrupt(void);
void cc77_rfs_disable_interrupt(void);
void cc77_pci_enable_interrupt(void);
void cc77_pci_clear_interrupt(void);
void cc77_crate_enable_lam(int lammask);
void cc77_crate_disable_lam(void);
void cc77_crate_z(int c);
void cc77_crate_c(int c);
void cc77_crate_seti(int c);
void cc77_crate_deli(int c);
void cc77_get_csrdata(int *ret);
void cc77_get_bmcsdata(int *ret);


static int done_try,lmsk;

struct pci_struct{
  int base1;
  int base2;
  unsigned int irq;
};


struct pci_struct pci_table[] ={
  {0, 0, 0}
};

struct pci_reg{
  unsigned int csr;
  unsigned int cnaf;
  unsigned int tcr;
  unsigned int srr;
  unsigned int fifo;
  unsigned int mwa;
  unsigned int mwtc;
  unsigned int mra;
  unsigned int mrtc;
  unsigned int ics;
  unsigned int bmcs;  
};

struct pci_reg reg[] ={
  {CSR, CNAF, TCR, SRR, FIFO, MWA, MWTC, MRA, MRTC, ICS, BMCS,},
};

void cc77_init_register(int irq,int ior,int pbor){

  pci_table->irq   = (unsigned int)irq;
  pci_table->base1 = (unsigned int)ior;
  pci_table->base2 = (unsigned int)pbor;
  
  reg->csr  += (unsigned int)pbor; /* for Create Control */
  reg->cnaf += (unsigned int)pbor; /* for CNAF */
  reg->tcr  += (unsigned int)pbor; /* Not use */
  reg->srr  += (unsigned int)pbor; /* for GO request */

  reg->fifo += (unsigned int)pbor;  /* for Data Transfer */
  reg->mwa  += (unsigned int)ior;  /* Not use */
  reg->mwtc += (unsigned int)ior;  /* Not use */
  reg->mra  += (unsigned int)ior;  /* Not use */
  reg->mrtc += (unsigned int)ior;  /* Not use */
  reg->ics  += (unsigned int)ior;  /* Not use */
  reg->bmcs += (unsigned int)ior;  /* Not use */

}

int cc77_check_done_retq(int try){
  int val;

  done_try = 0;
  while (!((val = inl(reg->csr)) & DONE)){
    done_try++;
    if(done_try > try) {
      /* printk("check_done: Timeout, Waiting for DONE\n"); */
      return -1;
      break;
    }
  }

  return val & NOQ;
}

int cc77_check_done(int try){
  int ret = 0;

  done_try = 0;
  while (!(inl(reg->csr) & DONE)){
    done_try++;
    if(done_try > try) {
      /* printk("check_done: Timeout, Waiting for DONE\n"); */
      ret = 1;
      break;
    }
  }
  return ret;
}

int cc77_check_emp_fifo(void){
  int ret = 0;

  return ret;
}

int cc77_check_full_fifo(void){
  int ret = 0;

  return ret;
}

int cc77_check_lam(void){
  return (inl(reg->cnaf) & lmsk) & 0x00ffffff;
}

void cc77_read_lam(int *data){
  *data = inl(reg->cnaf) & 0x00ffffff;
}

int cc77_control(int cnaf){
  outl(cnaf,reg->cnaf);
  outl((unsigned int)GO,reg->srr);
  return cc77_check_done_retq(MAX_DONE_TRY);
}

int cc77_read16(int cnaf,short *data){
  int ret;
  outl(cnaf,reg->cnaf);
  outl((unsigned int)GO,reg->srr);
  ret = cc77_check_done_retq(MAX_DONE_TRY);
  *data = inl(reg->fifo);
  return ret;
}

int cc77_read24(int cnaf,short *data){
  unsigned int tdata;
  int ret;
  outl(cnaf,reg->cnaf);
  outl((unsigned int)GO,reg->srr);
  ret = cc77_check_done_retq(MAX_DONE_TRY);
  tdata = inl(reg->fifo);
  *data = tdata & 0xffff;
  *(data+1) = (tdata>>16) & 0x00ff;
  return ret;
}

int cc77_write16(int cnaf,short *data){
  outl(cnaf,reg->cnaf);
  outl((unsigned int)*data,reg->fifo);
  outl((unsigned int)GO,reg->srr);
  return cc77_check_done_retq(MAX_DONE_TRY);
}

int cc77_write24(int cnaf,int *data){
  outl(cnaf,reg->cnaf);
  outl((unsigned int)*data,reg->fifo);
  outl((unsigned int)(SINGLE24 | GO),reg->csr);
  return cc77_check_done_retq(MAX_DONE_TRY);
}

int cc77_block_read16(int mode,int cnaf,short *data,int count){
  int ret = 0,btry = 0, br = 0;
  unsigned int tdata;

  outl(cnaf,reg->cnaf);

  while(1){
    outl((unsigned int)GO,reg->srr);
    if(mode == QREPEAT){
      if(cc77_check_done_retq(MAX_DONE_TRY) != 0){
	br = 1;
      }
    }else if(mode == QSTOP){
      if(cc77_check_done_retq(MAX_DONE_TRY) == 0){
	br = 1;
      }
    }else{
      if(cc77_check_done(MAX_DONE_TRY) == 1){
	/* printk("Error DONE in block_read16.\n"); */
	break;
      }
    }
    if(br == 0){
      tdata = inl(reg->fifo);
      *(data+btry) = tdata & 0xffff;
      btry++;
    }
    if(btry == count || br == 1){
      /* printk("btry = count\n"); */
      break;
    }
  }

  ret = btry;

  return ret;
}

/*
  int block_read24(int mode,int cnaf,short *data,int count){
  int ret = 0,btry = 0;
  unsigned int tdata;

  outl(cnaf,reg->cnaf);

  while(1){
  outl((unsigned int)GO,reg->srr);
  if(check_done(MAX_DONE_TRY) == 1){
  break;
  }
  tdata = inl(reg->fifo);
  *(data+btry) = tdata & 0xffff;
  btry++;
  *(data+btry) = (tdata >>16) & 0x00ff;
  btry++;
  if(btry == count*2){
  break;
  }
  }

  ret = btry;

  return ret;
  }
*/

int cc77_block_read24(int mode,int cnaf,unsigned int *data,int count){
  int ret = 0,btry = 0;

  outl(cnaf,reg->cnaf);

  while(1){
    outl((unsigned int)GO,reg->srr);
    if(cc77_check_done(MAX_DONE_TRY) == 1){
      /* printk("Error DONE in block_read24.\n"); */
      break;
    }
    *(data+btry) = inl(reg->fifo) & 0x00ffffff;
    btry++;
    if(btry == count){
      /* printk("btry = count\n"); */
      break;
    }
  }

  ret = btry;

  return ret;
}


int cc77_dma_block_read16(int mode,int cnaf,int count){
  int ret = 0;

  /* Not use */

  return ret;
}

int cc77_dma_block_read24(int mode,int cnaf,int count){
  int ret = 0;

  /* Not use */

  return ret;
}

int cc77_dma_block_write16(int mode,int cnaf,int count){
  int ret = 0;

  /* Not use */

  return ret;
}

void cc77_crate_reset(void){
  cc77_crate_z(0);
}

void cc77_rfs_enable_interrupt(void){
  outl(CC_EI,reg->csr);
}

void cc77_rfs_disable_interrupt(void){
  outl(0x00000000,reg->csr);
}

void cc77_pci_enable_interrupt(void){
  /* Not use */
}

void cc77_pci_clear_interrupt(void){
  /* Not use */
}

void cc77_crate_enable_lam(int lammask){
  lmsk = lammask & 0x00ffffff;
  //outl(CC_EI,reg->csr);
}
void cc77_crate_define_lam(int n){
  lmsk = (0x0001<<(n-1)) & 0x00ffffff;
}

void cc77_crate_disable_lam(void){
  outl(0x00000000,reg->csr);
}

void cc77_get_csrdata(int *ret){
  /* Not use */
}

void cc77_get_bmcsdata(int *ret){
  /* Not use */
}

void cc77_crate_z(int c){
  outl(CC_Z,reg->csr);
}

void cc77_crate_c(int c){
  outl(CC_C,reg->csr);
}

void cc77_crate_seti(int c){
  outl(CC_I,reg->csr);
}

void cc77_crate_deli(int c){
  outl(0x00000000,reg->csr);
}

int cc77_chkq(void){
  return inl(reg->csr) & NOQ;
}
